#!/bin/bash

# PURPOSE
# This is not a nightly script. You use it to regenerate the model-collections
# if Greenstone has changed fundamentally, such as what HASH OIDs get assigned
# to documents or something that changes the contents of the index and 
# archives folders. This has happened now with the commits 
# http://trac.greenstone.org/changeset/28022 and 
# http://trac.greenstone.org/changeset/28021
# These commits generate new stable HASH OIDs for the existing documents.


# USAGE
# Put this file in the toplevel of the Greenstone 2 binary/compiled SVN installation
# that you want to generate the model collections with.
# You can provide a list of collection names or none, in which case all the collections
# are processed.

# Pass in --svnupdate to copy across the contents of archives and index in the
# rebuilt collection, overwriting their equivalents in the svn model collection,
# but not removing any extraneous HASH folders already present.
# !!!!! IMPORTANT: if you pass in svnupdate, it leaves you to do the final commit on 
# the (svn) model-collect folder!

# Pass in --svndelete to remove the archives and index from svn in the model-collect
# and replace this with the rebuilt archives and index 
# The --svndelete is useful for when the HASH directory naming has changed and everything
# in archives and index has to be wiped out and moved back in from the rebuilt col.
# Passing in --svndelete will do the final commits on the model-collect folder.

# If neither flag is passed in, then the collections are rebuilt but the svn model-collect
# is not updated and the repository is not updated.

# Examples of usage:
# ./gen-model-colls.sh
# ./gen-model-colls.sh --svndelete
# ./gen-model-colls.sh --svnupdate Tudor-Basic Tudor-Enhanced

# The first just rebuilds all the collections in a new folder called collect and stops there

# The second rebuilds all the collections in collect and svn removes the archives and the index 
# folders in model-collect. Then it copies across the rebuilt archives and index into model-collect 
# and svn adds them.

# The third example checks out all the model-collections again, but rebuilds only the 2 collections 
# specified in the new collect folder. Then it copies across the *contents* of the archives and 
# index folders of those 2 collections into their model-collect equivalents. You then still have to 
# do the final svn commit on the model-collect folder after looking over the differences.

# Also valid examples:
# ./gen-model-colls.sh Tudor-Basic Tudor-Enhanced
# ./gen-model-colls.sh --svndelete Tudor-Basic Tudor-Enhanced
# ./gen-model-colls.sh --svnupdate

# PSEUDOCODE
# This script:
# Checks out the model-collections folder from SVN 
# Makes a copy
# In the copy: gets rid of their .svn folders, and builds each collection in turn, moving building to index once done
# If --svndelete was passed in: svn removes model-collect/archives and model-collect/index, copies over collect/index
# and collect/archives into model-collect and svn adds model-collect/archives and model-collect/index. Then SVN COMMITS
# model-collect/archives and model-collect/index.
# If --svnupdate was passed in: copies collect/archives/* into model-collect/archives/*, and copies collect/index/*
# into model-collect/index/*, overwriting files that already existed but have now been updated upon rebuild. However, 
# --svnupdate will leave untouched any files and folders unique to model-collect. No SVN commit, that's LEFT UP TO YOU.

# See earlier version of this script:
# To svn remove what's unique to model-collect and svn add what's been rebuilt in index and archives
# see http://stackoverflow.com/questions/7502261/delete-folder-content-and-remove-from-version-control

# http://stackoverflow.com/questions/5044214/how-do-i-detect-and-or-delete-empty-subversion-directories
# http://stackoverflow.com/questions/1301203/removing-svn-files-from-all-directories

#*******************************GLOBAL VARIABLES***************************

# mode can be svndelete or svnupdate
mode=
debug_mode=0
commit_message=

#*****************************FUNCTIONS*****************************

# DON'T ADD ANY FURTHER ECHO STATEMENTS IN FUNCTION get_col_basename
# "you have to be really careful on what you have in this function, as having any code which will eventually echo will mean that you get incorrect return string."
# see http://stackoverflow.com/questions/3236871/how-to-return-a-string-value-from-a-bash-function
function get_col_basename () {
    collection=$1

    #escape the filename (in case of space)
    collection=`echo $collection | sed 's@ @\\\ @g'`
    
    #get just the basename
    collection=`basename $collection`

    # returning a string does not work in bash
    # see http://stackoverflow.com/questions/3236871/how-to-return-a-string-value-from-a-bash-function    

    #return $collection
    echo $collection
}


# Function that handles the --svndelete flag (mode) of this script for a single collection
function svn_delete () {

    # svn remove archives and index in each collection
    # commit them all
    # copy over newly rebuilt archives and index into each model-collection
    # svn add the new archives and index folders of each collection
    # commit them all
    
    
    if [ "x$1" == "x" ]; then
	for collection in collect/*; do	    
	    _del_col_archives_index $collection
	done
    else 
	for collection in "$@"; do
	    _del_col_archives_index $collection
	done
    fi

    # svn commit all the svn rm statements done above in one go:
    # don't do `svn up` at this point, as doing so will then retrieve all the folders that just were svn-removed

    if [ "x$commit_message" == "x" ]; then
	commit_message="Clean rebuild of model collections 1/2. Clearing out deprecated archives and index."
    fi

    # Numerical comparisons: http://tldp.org/LDP/abs/html/comparison-ops.html
    if [ "$debug_mode" -eq "0" ]; then
	svn commit -m "AUTOCOMMIT by gen-model-colls.sh script. Message: $commit_message" model-collect
    fi

    # Having svn committed the deletes, do an svn up to locally delete what was svn-removed above, 
    # BEFORE copying from the rebuilt archives and index folders
    if [ "$debug_mode" -eq "0" ]; then
	svn up model-collect	
    fi

    # copy from the rebuilt archives and index over into the svn model-collect and svn add them
    if [ "x$1" == "x" ]; then
	for collection in collect/*; do	    
	    _add_col_archives_index $collection
	done
    else 
	for collection in "$@"; do
	    _add_col_archives_index $collection
	done
    fi

    # commit all the svn add statements done just above in one go
    if [ "x$commit_message" == "x" ]; then
	commit_message="Clean rebuild of model collections 2/2. Adding rebuilt archives and index."
    fi

    if [ "$debug_mode" -eq "0" ]; then
	svn commit -m "AUTOCOMMIT by gen-model-colls.sh script. Message: $commit_message" model-collect
    fi

    echo
    echo "*********************"
    echo "Done svn-deleting rebuilt model-collection: $collection"
    echo "*********************"
    echo
}

# To undo the changes made by svndelete, run the following manually
# svn revert --depth infinity <model-collect/$collection/archives/*
# svn revert --depth infinity <model-collect/$collection/archives/*
# then remove both the local archives and index, and do an svn up to get original checkout back

# svn delete this collection's archives and index folders
# (The commit will be done when in one step for all collections on which this function was called)
function _del_col_archives_index () {
    collection=$1

    #escape the filename (in case of space) and get just the basename
    collection=$(get_col_basename $collection)

    if [ ! -e model-collect/$collection ]; then
	echo "del_col_archives_index: $collection does not exist in model-collect"
	return;
    fi

    # remove the entire archives and index folders from svn
    if [ "$debug_mode" -eq "0" ]; then
	svn rm model-collect/$collection/archives
	svn rm model-collect/$collection/index
    elif [ "$debug_mode" -eq "1" ]; then
	rm -rf model-collect/$collection/archives
	rm -rf model-collect/$collection/index
    fi

}


# copy and then svn add the collection's archives and index folders
function _add_col_archives_index () {
    collection=$1

    #escape the filename (in case of space) and get just the basename
    collection=$(get_col_basename $collection)

    if [ ! -e model-collect/$collection ]; then
	echo "add_col_archives_index: $collection does not exist in model-collect"
	return;
    fi

    # remove the entire archives and index folders from svn
    cp -r collect/$collection/archives model-collect/$collection/.
    cp -r collect/$collection/index model-collect/$collection/.	

    if [ "$debug_mode" -eq "0" ]; then
	svn add model-collect/$collection/archives
	svn add model-collect/$collection/index
    fi
}


# UNUSED, but useful for spotting differences between the collect and model-collect
# after rebuild, before svn updating/deleting, as opposed to at the end of the script
function svn_process_single_collection () {
    collection=$1
    
    #escape the filename (in case of space) and get just the basename
    collection=$(get_col_basename $collection)

    if [ ! -e model-collect/$collection ]; then
	echo "svn_process_single_collection: $collection does not exist in model-collect"
	return;
    fi

# return here if just deleting empty dirs
#return

    # diff the svn model and rebuilt model collections
    diff_result=`diff -rq model-collect/$collection collect/$collection | grep -v ".svn"`
#    echo "Diff result for collection $collection: $diff_result"

    # if no differences in the current collection, then we're done
    if [ "x$diff_result" == "x" ]; then
	echo "No differences in collection $collection"
	return;
    fi

    # check that none of the lines mention files outside the archives or index folders
    # http://en.gibney.org/tell_the_bash_to_split_by_newline_charac
    # http://forums.gentoo.org/viewtopic-p-3130541.html

    # http://wi-fizzle.com/article/276
    # http://stackoverflow.com/questions/918886/how-do-i-split-a-string-on-a-delimiter-in-bash
    # http://www.linuxquestions.org/questions/programming-9/split-a-string-on-newlines-bash-313206/
    # http://unix.stackexchange.com/questions/39473/command-substitution-splitting-on-newline-but-not-space

    # store backup of Internal Field Separator value, then set IFS to newline for splitting on newline

    IFS_BAK=$IFS
#    IFS='\n' # splits on all whitespace
IFS='
'
    # in the lines returned from the diff, test for archives or newline
    # http://stackoverflow.com/questions/229551/string-contains-in-bash
    for line in `diff -rq model-collect/$collection collect/$collection | grep -v ".svn"`; do
	# echo "LINE: $line"	
	if [[ "$line" != *archives* && "$line" != *index* ]]; then
	    # the file that is different is neither in index nor in archives, send this diffline to the report
	    echo $line >> report.txt
	fi
    done
 
    IFS=$IFS_BAK
    IFS_BAK=
}

# Function that takes care of the --svnupdate flag mode of this script for a single collection
function update_single_collection () {
    collection=$1
    
    #escape the filename (in case of space) and get just the basename
    collection=$(get_col_basename $collection)

    if [ ! -e model-collect/$collection ]; then
	echo "update_single_collection: $collection does not exist in model-collect"
	return;
    fi

    # copy across the contents of the rebuilt model-collection's index and archives to the svn model-collect
    cp -r collect/$collection/archives/* model-collect/$collection/archives/.
    cp -r collect/$collection/index/* model-collect/$collection/index/.

    # now svn add any and all the NEW items in model-collect's archives and index
    # see http://stackoverflow.com/questions/1071857/how-do-i-svn-add-all-unversioned-files-to-svn
#    if [ "$debug_mode" -eq "0" ]; then
	svn add --force model-collect/$collection/archives/* --auto-props --parents --depth infinity -q
	svn add --force model-collect/$collection/index/* --auto-props --parents --depth infinity -q
#    fi

    echo "svn model-collect update process complete. CHECK AND COMMIT THE model-collect FOLDER!"

    # if etc/collect.cfg is different, copy it across too?

    echo
    echo "*********************"
    echo "Done updating the rebuilt LOCAL model-collection: model-collect/$collection"
    echo "*********************"
    echo
}


# re-build a single collection in "collect" which is a copy of model-collect
function build_single_collection () {
    collection=$1
   
    collection=$(get_col_basename $collection)
    
    import.pl -removeold $collection
    buildcol.pl -removeold $collection
    rm -rf collect/$collection/index
    mv collect/$collection/building collect/$collection/index
    
    echo
    echo "*********************"
    echo "Done rebuilding model collection: $collection"
    echo "*********************"
    echo
}


# http://stackoverflow.com/questions/16483119/example-of-how-to-use-getopt-in-bash
function usage() { 
# usage() { echo "Usage: $0 [-s <45|90>] [-p <string>]" 1>&2; exit 1; }

    echo "*******************************************"
    echo "Usage: $0 [--svnupdate|--svndelete] [col1, col2, col3,...]"; 
    echo "If no collections are provided, all collections will be processed.";
    echo "If neither svnupdate nor svndelete are provided, svnupdate is assumed.";
    echo "*******************************************"
    exit 1; 
}


#*******************************MAIN PROGRAM***************************

# process optional command line arguments
# http://blog.onetechnical.com/2012/07/16/bash-getopt-versus-getopts/
# Execute getopt
ARGS=$(getopt -o m:uxd -l "message:,svnupdate,svndelete,debug" -n "$0" -- "$@");

#Bad arguments
if [ $? -ne 0 ];then
    usage
    exit 1
fi

eval set -- "$ARGS";


# -n: http://tldp.org/LDP/abs/html/testconstructs.html
while true; do
  case "$1" in
    -x|--svndelete)
      shift;
      if [ "x$mode" == "xsvnupdate" ]; then
	  echo
	  echo "Can't use both svndelete and svnupdate"
	  usage
	  exit 1
      else
	  mode=svndelete
      fi
      ;;
    -u|--svnupdate)
      shift;
      if [ "x$mode" == "xsvndelete" ]; then
	  echo
	  echo "Can't use both svndelete and svnupdate"
	  usage
	  exit 1
      else
	  mode=svnupdate
      fi
      ;;
    -d|--debug)
      shift;
      debug_mode=1
      ;;
    -m|--message)
      shift;
      if [ -n "$1" ]; then
	  commit_message=$1
          shift;
      fi
      ;;
    --)
      shift;
      break;
      ;;
  esac
done

#echo "commit message: $commit_message"
#echo "Debug mode is: $debug_mode"
#exit


# If no mode provided (svndelete|svnupdate) as cmd line arg, then don't modify
# the svn model-collect folder. Then this script stops after rebuilding the model-copy in collect

# the remaining arguments to the script are assumed to be collections

# debugging
#for collection in "$@"; do
#    collection=collect/$collection
#    echo "Collection: $collection"
#done

# finished processing arguments


# report will contain the output of the diff for
if [ -f report.txt ]; then
    rm report.txt
fi

# Need pdfbox for the PDFBox tutorial
if [ ! -e ext/pdf-box ]; then
    cd ext
    if [ ! -e ext/pdf-box-java.tar.gz ]; then
	wget http://trac.greenstone.org/export/head/gs2-extensions/pdf-box/trunk/pdf-box-java.tar.gz
	tar -xvzf pdf-box-java.tar.gz
    fi
    cd ..
fi


# move the existing collect folder out of the way
if [ -e collect ] && [ ! -e collect_orig ] ; then
    mv collect collect_orig
fi


# get model-collect from svn
# if we already have it, svn update the entire model-collect folder if processing all collections
# or svn update just any collections specified in the model-collect folder
if test -e model-collect; then
    if [ "$1" == "" ]; then
	svn up model-collect
    else
	for collection in "$@"; do
	    svn up model-collect/$collection
	done
    fi
else
    svn co http://svn.greenstone.org/other-projects/nightly-tasks/diffcol/trunk/model-collect
fi

# Make a copy of the model-collect named as the new collect 
# (or if collections are specified in the cmdline arguments, copy just these over from model-collect into collect)
# Then remove the copy's .svn folders
echo "***********************************************"
echo "Creating a copy of the model-collect folder as folder collect and removing the .svn subfolders from the copy:"
echo
if [ -e collect_orig ]; then
    if [ ! -e collect ]; then
	cp -r model-collect collect	
	find collect -name ".svn" -type d -exec rm -rf {} \; #2>&1 > /dev/null
    else
	if [ "$1" == "" ]; then
	    rm -rf collect
	    cp -r model-collect collect
	    find collect -name ".svn" -type d -exec rm -rf {} \;
	else
	    for collection in "$@"; do
		if [ -e collect/$collection ]; then
		    rm -rf collect/$collection
		fi
		cp -r model-collect/$collection collect/$collection
		find collect/$collection -name ".svn" -type d -exec rm -rf {} \;
	    done
	fi
    fi
fi
echo "***********************************************"

# Set up the Greenstone environment for building
source setup.bash

# parse arguments
# http://stackoverflow.com/questions/12711786/bash-convert-command-line-arguments-into-array
# http://stackoverflow.com/questions/255898/how-to-iterate-over-arguments-in-bash-script

if [ "$1" == "" ]; then

    # all_collections	
    #for each collection, import, build, move building to index
    for collection in collect/*; do
	build_single_collection $collection;

	if [ "x$mode" != "x" ]; then
	    #svn_process_single_collection $collection

	    if [ "x$mode" == "xsvnupdate" ]; then
		update_single_collection $collection
	    fi
	fi
    done

    # having rebuilt all the collections, just the processing for svndelete/update remains:
    if [ "x$mode" == "xsvndelete" ]; then
	svn_delete
    fi

else
    # Command-line args are a list of collections, 
    # process each command-line arg, after confirming such a collection exists

    for collection in "$@"; do
	collection=collect/$collection
	if test -e $collection; then   
	    build_single_collection $collection;

	    if [ "x$mode" != "x" ]; then
		#svn_process_single_collection $collection

		if [ "x$mode" == "xsvnupdate" ]; then
		    update_single_collection $collection
		fi
	    fi
	else
	    echo "Can't find collection $collection. Skipping."
	fi
    done

    # having rebuilt the specified collections above, just the processing for svndelete/update remains
    if [ "x$mode" == "xsvndelete" ]; then
	svn_delete $@
    fi

fi


echo
echo "*****************************************"
echo
# NO LONGER NECESSARY: WE'RE DOING A DIFF BETWEEN collect AND model-collect AT THIS SCRIPT'S END
# if we were svn updating/deleting collections, then mode was set
# if in that case a report was generated with additional differences, point the user to it
#if [ -f report.txt ] && [ "x$mode" != "x" ]; then
#    echo "Some files or folders outside of archives and index directories were different. See report.txt"
#    echo
#fi

# if not svnupdating or svndeleting, then inform the user that model-collect is unchanged
# if svnupdating, then warn the user that model-collect still needs committing
# if svndeleting, then inform the user that model-collect has been changed and committed
if [ "x$mode" == "x" ]; then
    echo "* The model-collect folder has not been altered. Changes have only been made to collect"
elif [ "x$mode" == "xsvnupdate" ]; then
    echo "* TO DO: You still need to run svn status and then svn commit on the model-collect folder. Besides that:"    
elif [ "x$mode" == "xsvndelete" ]; then
    echo "* The model-collect folder's archives and index subfolders have been updated and committed to svn."
fi
echo

if [ "x$mode" != "x" ]; then
    echo "* DIFFERENCES REMAINING BETWEEN model-collect AND collect (skipping .svn folders):"
    echo
    if [ "$1" == "" ]; then
	echo "---START DIFF---"
	diff -rq model-collect collect | grep -v ".svn"
    else
	for collection in "$@"; do
	    echo "--COLLECTION: $collection"
	    diff -rq model-collect/$collection collect/$collection | grep -v ".svn"
	    echo "--"
	done
    fi
    echo "---END DIFF---"
    echo
fi

echo "* The original collect directory has been left renamed as collect_orig"
echo

if [ "$debug_mode" -eq "1" ]; then
    echo "* This script was run in DEBUG MODE, nothing has been changed in svn"
fi
echo
echo "*****************************************"
echo


# deletes empty dirs
#	    find collect/$collection/archives/HASH* -type d -empty -delete
#	    find collect/$collection/index/assoc/HASH* -type d -empty -delete

# To recursively delete all empty dirs in the copy of model-collect (since the dirs will not have .svn folders in them anymore)
# http://www.commandlinefu.com/commands/view/5131/recursively-remove-all-empty-directories
#find collect -type d -empty -delete

# The following when put in a separate script file will delete all folders from model-collect that are
# empty in the copied collection (all folders which contain only a .svn subfolder in model-collect)
# ---------------------------------------------
#!/bin/bash

#for collection in collect/*; do
    #escape the filename (in case of space)
#    collection=`echo $collection | sed 's@ @\\\ @g'`
    
    #get just the basename
#    collection=`basename $collection`

    # HASH dirs that are empty in local collect's archives and index/assoc, 
    # need to be removed from the svn in model-collect

#    for line in `find collect/$collection/archives/HASH* -type d -empty`; do
#	modelline="model-$line"
#	echo "LINE: $modelline"
	
	# remove from svn of model collect
#	svn rm $modelline
##	rm -rf $modelline
	# remove physically from local collect
#	rm -rf $line
#    done

#    for line in `find collect/$collection/index/assoc/HASH* -type d -empty`; do
#	modelline="model-$line"
#	echo "LINE: $modelline"
	
	# remove from svn of model collect
#	svn rm $modelline
##	rm -rf $modelline
	# remove physically from local collect
#	rm -rf $line
#    done

#done
# ---------------------------------------------
